/*****************************************************************************/
/* BroadVoice(R)32 (BV32) Floating-Point ANSI-C Source Code                  */
/* Revision Date: August 19, 2009                                            */
/* Version 1.0                                                               */
/*****************************************************************************/

/*****************************************************************************/
/* Copyright 2000-2009 Broadcom Corporation                                  */
/*                                                                           */
/* This software is provided under the GNU Lesser General Public License,    */
/* version 2.1, as published by the Free Software Foundation ("LGPL").       */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY SUPPORT OR WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the LGPL for     */
/* more details.  A copy of the LGPL is available at                         */
/* http://www.broadcom.com/licenses/LGPLv2.1.php,                            */
/* or by writing to the Free Software Foundation, Inc.,                      */
/* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 */
/*****************************************************************************/


/*****************************************************************************
  lspdec.c :  LSP decoding function

  $Log$
******************************************************************************/

#include <stdio.h>
#include "typedef.h"
#include "bv32externs.h"
#include "bvcommon.h"

void vqdec(Float *, short, Float *, int);

void lspdec(
            Float   *lspq,  
            short   *lspidx,  
            Float   *lsppm,
            Float	*lspq_last)
{
   Float elsp[LPCO], lspe[LPCO]; 
   Float lspeq1[LPCO], lspeq2[LPCO];
   Float a0, *fp1, *fp2;
   int i, k, lsfdordr;
   
   /* CALCULATE ESTIMATED (MA-PREDICTED) LSP VECTOR */
   fp1 = lspp;
   fp2 = lsppm;
   for (i = 0; i < LPCO; i++) {
      a0 = 0.0F;
      for (k = 0; k < LSPPORDER; k++) {
         a0 += *fp1++ * *fp2++;
      }
      elsp[i] = a0;
   }
   
   /* PERFORM FIRST-STAGE VQ CODEBOOK DECODE */
   vqdec(lspeq1,lspidx[0],lspecb1,LPCO);
   
   
   /* PERFORM SECOND-STAGE VQ CODEBOOK DECODE */
   vqdec(lspeq2,lspidx[1],lspecb21,SVD1);
   vqdec(lspeq2+SVD1,lspidx[2],lspecb22,SVD2);
   
   /* GET OVERALL QUANTIZER OUTPUT VECTOR OF THE TWO-STAGE VQ */
   /* AND CALCULATE QUANTIZED LSP */
   for (i = 0; i < LPCO; i++) {
      lspe[i] = lspeq1[i] + lspeq2[i];
      lspq[i] = lspe[i] + elsp[i] + lspmean[i];
   }
   
   /* detect bit-errors based on ordering property */
   if (lspq[0] < 0.0)
      lsfdordr = 1;
   else
      lsfdordr = 0;
   for (i=1; i<SVD1; i++) {
      if(lspq[i] - lspq[i-1] < 0.0) lsfdordr = 1;
   }
   
   /* substitute LSP and MA predictor update if bit-error detected */
   if(lsfdordr) {
      for(i=0; i<LPCO; i++){
         lspq[i] = lspq_last[i];
         lspe[i] = lspq[i] - elsp[i] - lspmean[i];
      }
   }
   
   /* UPDATE LSP MA PREDICTOR MEMORY */
   i = LPCO * LSPPORDER - 1;
   fp1 = &lsppm[i];
   fp2 = &lsppm[i - 1];
   for (i = LPCO - 1; i >= 0; i--) {
      for (k = LSPPORDER; k > 1; k--) {
         *fp1-- = *fp2--;
      }
      *fp1-- = lspe[i];
      fp2--;
   }
   
   
   /* ENSURE CORRECT ORDERING & MINIMUM SPACING TO GUARANTEE STABILITY */
   stblz_lsp(lspq, LPCO);
   
}


void vqdec(
           Float   *xq,    /* VQ output vector (quantized version of input vector) */
           short   idx,    /* VQ codebook index for the nearest neighbor */
           Float   *cb,    /* VQ codebook */
           int     vdim    /* vector dimension */
           )
{
   int j, k;
   
   j = idx * vdim;
   for (k = 0; k < vdim; k++) {
      xq[k] = cb[j + k];
   }
}  


void lspplc(
            Float   *lspq,  
            Float   *lsppm)
{
   Float elsp[LPCO];
   Float a0, *fp1, *fp2;
   int i, k;
   
   /* CALCULATE ESTIMATED (MA-PREDICTED) LSP VECTOR */
   fp1 = lspp;
   fp2 = lsppm;
   for (i = 0; i < LPCO; i++) {
      a0 = 0.0F;
      for (k = 0; k < LSPPORDER; k++) {
         a0 += *fp1++ * *fp2++;
      }
      elsp[i] = a0;
   }
   
   /* UPDATE LSP MA PREDICTOR MEMORY */
   i = LPCO * LSPPORDER - 1;
   fp1 = &lsppm[i];
   fp2 = &lsppm[i - 1];
   for (i = LPCO - 1; i >= 0; i--) {
      for (k = LSPPORDER; k > 1; k--) { *fp1-- = *fp2--; }
      *fp1-- = lspq[i] - lspmean[i] - elsp[i];
      fp2--;
   }
}
